# History: Jul 19 2011 Add error message in find.partition if p-vsalue is too large

## t.obs is the observed test statistic

SAMC.adapt<-function(data.input, t.obs, t.start=0, n.iter.1=200000, n.iter.2=1000000, 
         n.region.1=101, n.region.2=301,
         prop.change=0.05, gain.factor.t0=1000,
         fun.test.statistic=NULL, fun.proposal=NULL, ...)
{

    if (!is.function(fun.test.statistic)) stop("ERROR: fun.test.statistic must be a function")
    if (!is.function(fun.proposal)) stop("ERROR: fun.proposal must be a function")
    ## n.region.1 include the last subregion
    n.cut<-n.region.1 - 1
    t.end<-t.obs
    t.step<-(t.end - t.start)/n.cut
    partition.vec.1 <-t.start + t.step * c(0:n.cut)
    partition.vec.1[n.cut+1]<-t.end


    theta.start.1<-rep(0, length(partition.vec.1))

    res.1<-SAMC.fun(data.input, n.iter.1, partition.vec.1, prop.change,
	 theta.start.1, gain.factor.t0, fun.test.statistic, fun.proposal)
      
    if (n.iter.2 > 0)
    {
       hist.mat.1<-res.1$hist.mat[,c(1,4)]
       out.1<-find.partition(hist.mat=hist.mat.1, n.interval=n.region.2, 
            my.precision=1.0, num.tail=1)

       partition.vec.2<-out.1[,2]
       theta.start.2<-out.1[,1]

       res.1<-SAMC.fun(data.input, n.iter.2, partition.vec.2, prop.change,
	 theta.start.2, gain.factor.t0, fun.test.statistic, fun.proposal, ...)
    }
    res.1
}

## switch between two vectors sequentially
seq.switch<-function(vec.1, vec.2, n.switch)
{
   i.1.vec <- sample(x=1:length(vec.1), size=n.switch, replace=T)
   i.2.vec <- sample(x=1:length(vec.2), size=n.switch, replace=T)
     
   i.1.vec <- as.integer(i.1.vec - 1)
   i.2.vec <- as.integer(i.2.vec - 1)
   temp    <- .C("seq_switch", i.1.vec, i.2.vec, as.integer(n.switch), vec1=as.double(vec.1), vec2=as.double(vec.2))
   v.1.new <- temp$vec1
   v.2.new <- temp$vec2
 
   list(v.1.new=v.1.new, v.2.new=v.2.new)
}

seq.switch.all<-function(vec.1, vec.2, n.switch, use.C.code=1)
{
   i.1.vec <- sample(x=1:length(vec.1), size=n.switch, replace=T)
   i.2.vec <- sample(x=1:length(vec.2), size=n.switch, replace=T)
    
   if (use.C.code) {
     i.1.vec <- as.integer(i.1.vec - 1)
     i.2.vec <- as.integer(i.2.vec - 1)
     temp    <- .C("seq_switch", i.1.vec, i.2.vec, as.integer(n.switch), vec1=as.double(vec.1), vec2=as.double(vec.2))
     v.1.new <- temp$vec1
     v.2.new <- temp$vec2
   } else {
     v.1.new <-vec.1
     v.2.new <-vec.2
     for (jj in 1:n.switch)
     {
        select.1<-v.1.new[i.1.vec[jj]]
        select.2<-v.2.new[i.2.vec[jj]]
        v.1.new[i.1.vec[jj]]<-select.2
        v.2.new[i.2.vec[jj]]<-select.1
     }
   }

   list(v.1.new=v.1.new, v.2.new=v.2.new)
}



## An t-test example fro the fun.test.statistic
t.test.statistic<-function(data.input)
{
   x<-data.input$x
   y<-data.input$y
   abs(t.test(x=x, y=y, var.equal=T)$statistic)
}

## a MH proposal to generate new data based on the current data, an example for permutation
proposal.permute.vector<-function(data.input, prop.change)
{ 
  x<-data.input$x
  y<-data.input$y
 
  n.x<-length(x)
  n.y<-length(y)
  n.switch<-floor(prop.change * min(n.x,n.y))

  vec.1<-c(1:n.x)
  vec.2<-c(1:n.y)+n.x
  res<-seq.switch(vec.1, vec.2, n.switch)
  new.vec.1<-res$v.1.new
  new.vec.2<-res$v.2.new

  ## reassign
  data.pool<-c(x,y)
  x.new<-data.pool[new.vec.1]
  y.new<-data.pool[new.vec.2]

  list(x=x.new, y=y.new)
}



TREND.TEST <- function(case.geno, ctrl.geno, phi=c(0,1/2,1))
{
    case.geno <- case.geno[!is.na(case.geno)]
    ctrl.geno <- ctrl.geno[!is.na(ctrl.geno)]
    ri <- c(sum(case.geno==0), sum(case.geno==1), sum(case.geno==2))
    si <- c(sum(ctrl.geno==0), sum(ctrl.geno==1), sum(ctrl.geno==2))
    ni <- ri + si                                                           
    r <- sum(ri)
    S <- sum(si)

    n <- r + S
    
    numerator <- sqrt(n) * sum(phi*(S*ri-r*si))

    denominator <- sqrt(r*S*(n*sum(phi^2*ni)-(sum(phi*ni))^2))

    temp <- numerator/denominator
    abs(temp)

   #  2*pnorm(-abs(temp))    

}

## trend test
trend.test.statistic<-function(data.input)
{
   case.geno<-data.input$x
   ctrl.geno<-data.input$y
   phi<-c(0, 0.5, 1)	
   TREND.TEST(case.geno, ctrl.geno, phi)
}

## missing is NA, genotype is 0, 1, or 2
MAX.test.statistic<-function(data.input)
{
   geno.ca<-data.input$x
   geno.cn<-data.input$y

   n.snp<-dim(geno.ca)[2]
   phi<-c(0, 0.5, 1)

   t.vec<-rep(-9, n.snp)
   
   for (i in 1:n.snp)
   {
     t.vec[i]<-TREND.TEST(geno.ca[,i], geno.cn[,i], phi)
   }

  t.sort<-sort(t.vec, decreasing=T)
  t.sort[1]
}


proposal.permute.matrix<-function(data.input, prop.change)
{ 
  x<-data.input$x
  y<-data.input$y

  n.x<-dim(x)[1]
  n.y<-dim(y)[1]
  n.switch<-floor(prop.change * min(n.x,n.y))
 
  vec.1<-c(1:n.x)
  vec.2<-c(1:n.y)+n.x
  res<-seq.switch(vec.1, vec.2, n.switch)
  new.vec.1<-res$v.1.new
  new.vec.2<-res$v.2.new

  ## reassign
  data.pool<-rbind(x,y)
  x.new<-data.pool[new.vec.1,]
  y.new<-data.pool[new.vec.2,]

  list(x=x.new, y=y.new)
}


## assuming the last partition.vec element is the obseved statistic
SAMC.fun<-function(data.input, n.iter, partition.vec, prop.change,
	 theta.start, gain.factor.t0, fun.test.statistic, fun.proposal, ...)
{
  t.0 <- fun.test.statistic(data.input, ...)
  t.start<-partition.vec[1]
 
  if (t.0 < t.start)
  { 
     stop("the starting point is too small \n")
  }   

  n.interval<-length(partition.vec)
  hist.mat<-matrix(0, ncol=3, nrow=n.interval)
  hist.mat[,1]<-partition.vec
  hist.mat[,3]<-0
  hist.mat[,2]<-theta.start
  pi.vec<-rep(1/n.interval, n.interval)

  t.1<-t.0
  k.1<-sum(partition.vec <= t.1)

  data.current<-data.input
  for (i in 1:n.iter)
  {
    ## generate a new data based on the proposal
    data.new <-fun.proposal(data.input=data.current, prop.change=prop.change, ...)
 
    t.2<-fun.test.statistic(data.new, ...)
    
    ## get the index for t.1 and t.2

    k.2<-sum(partition.vec <= t.2)

    ## check whether to update
    r<-hist.mat[k.1,2] - hist.mat[k.2,2]

    ## here assume the accept prob only depend on the theta difference, as it is usually the case
    prob.r<-min(1, exp(r))
    i.sample<-sample(x=c(0,1), size=1, prob=c(prob.r, (1-prob.r)))
    if (i.sample == 0)
    {
      ## accept
      t.1<-t.2
      k.1<-k.2
      data.current <-data.new
    }
    ## no need to change if not accepted
    ### update hist.mat
    hist.mat[k.1,3]<-hist.mat[k.1,3]+1
    my.e<-rep(0, n.interval)
    my.e[k.1]<-1
    r.t<-gain.factor.t0 / max(gain.factor.t0, i)
    hist.mat[,2]<-hist.mat[,2] + r.t * (my.e - pi.vec)
  }

  ## rescale the weight
  n.empty<-sum(hist.mat[,3]==0)
  d<-sum(pi.vec[hist.mat[,3]==0])/(n.interval - n.empty)
  my.adj<-hist.mat[,2] + log(pi.vec + d)
  maxw<-max(my.adj)
  my.w<-exp(my.adj-maxw)/sum(exp(my.adj-maxw))
  my.w[hist.mat[,3]==0]<-0
  my.w<-my.w/sum(my.w)  
  hist.mat<-data.frame(hist.mat, my.w)
  names(hist.mat)<-c("partition", "theta", "samplefreq", "weight")

  ## get the p-value
  p.value<-my.w[n.interval]
  list(hist.mat=hist.mat, p.value = p.value)
}


## p.end is the samllest prob. unit
# n.interval: is the number of interval from 0-(1-p.remain)
# there is n.interval grid point, convering the p.remain

grid.define<-function(p.end, n.interval, p.remain)
{
   t<-p.end
   n<-n.interval-1
   delta<-1- p.remain
   my.fun<-function(x)
   {
      t/(x^(n-1)) * (1-x^n) / (1-x) - delta
   } 
   res<-uniroot(my.fun, lower=0.01, upper=0.999,tol = 1e-30,  maxiter=10000)

   r<-res$root
   a<-p.end/ (r^(n-1))
   v<-a*(r^(c(0:(n-1))))
   
   ## the length is n+1
   v<-c(v, p.remain)
   v<-v/sum(v)
   v.2<-cumsum(rev(v))
   v.2
}



## the first colunm is the cut point, a1-a2-a3-a4-+
# my.precision is the precisoin relative to the estimated tail probabilty
# num.tail is the number of unit measrued by the tail probabilty used in the
# second part, including the last interval

## if num.tail = 1, then the num of region is n.interval
## if num.tail > 1, then the num of region in (num.tail-1)/my.precision + n.interval

find.partition<-function(hist.mat, n.interval, my.precision=1.0, num.tail=1)
{
    my.t<-hist.mat[,1]
    my.p<-hist.mat[,2]

    p.2<-cumsum(rev(my.p)) ## from small to large
    t.2<-rev(my.t)

    n.1<-length(my.t)
    
    t.last<-my.t[n.1]
    p.tail.last<-my.p[n.1]

    p.end.unit<-p.tail.last * my.precision
    p.remain<-num.tail * p.tail.last
       
    ### check to make sure the tail prob is not too large

    ck.p<-n.interval * p.end.unit
    if (ck.p > (1-p.remain))
    {
       stop("ERROR: the p-value is too large, please just use the standard permutation,
              the program should only be used for evaluate p-value less than 10^-3 \n")
    }
       
    p.target<-grid.define(p.end.unit, n.interval, p.remain)
    p.vec<-p.target
  
    my.fun1<-approxfun(x=t.2, y=p.2, rule=2)

    my.fun2<-function(x,q)
    {
        my.fun1(x)-q
    }

    nn<-length(p.vec)
    x.vec<-rep(0, nn)
  
    num.ex.tail<-floor((num.tail-1) / my.precision)

    if (num.ex.tail > 0)
    {
      ## only when num.ex.tail >0, we need to find x.vec[1]
      x.vec[1]<-uniroot(f=my.fun2, q=p.vec[1], lower=0, 
               upper=t.last, tol=1e-30, maxiter=10000)$root
  
      for (i in 2:(nn-1))
      {
        x.vec[i]<-uniroot(f=my.fun2, q=p.vec[i], lower=0, 
               upper=x.vec[i-1], tol=1e-30, maxiter=10000)$root
      }

      ## in this part, the p and x are evenly distributed

      d.1<-x.vec[1]
      d.2<-t.last
    
      d.3<-(d.2-d.1)/num.ex.tail
      x.2<-d.1 + d.3 * c(0:num.ex.tail)
      x.2[num.ex.tail+1]<-d.2
      x.2<-rev(x.2[-1])
      x.final<-c(x.2, x.vec)
      theta.x.vec<-p.vec[2:nn] - p.vec[1:(nn-1)]
      theta.x.tail<-rep(p.end.unit, length(x.2))
      theta.vec<-c(p.tail.last, theta.x.tail, theta.x.vec)
    }
    if (num.ex.tail == 0)
    {
      x.vec[1]<-t.last
      for (i in 2:(nn-1))
      {
        x.vec[i]<-uniroot(f=my.fun2, q=p.vec[i], lower=0, 
               upper=x.vec[i-1], tol=1e-30, maxiter=10000)$root
      }
      x.final<-x.vec
      theta.x.vec<-p.vec[2:nn] - p.vec[1:(nn-1)]
      theta.vec<-c(p.tail.last, theta.x.vec)
    }
    cbind(rev(log(theta.vec)), rev(x.final))
}

## simulate genotype on cases and control, rr = (1, OR1, OR2)  
gen.data<-function(n.case, n.control, maf, rr)
{
   geno.freq<-c((1-maf) * (1-maf), 2*maf*(1-maf), maf*maf)

   geno.freq.case<-geno.freq * rr
   geno.freq.case<-geno.freq.case / sum(geno.freq.case)

  
   sample.control<-sample(x=c(0, 1, 2), size=n.control, replace=T, prob=geno.freq)
   sample.case<-sample(x=c(0, 1, 2), size=n.case, replace=T, prob=geno.freq.case)

   list(x=sample.control, y=sample.case)
}

